from importlib import import_module
from .exception import FrictionlessException
from .errors import PipelineError, TaskError
from .status import Status, StatusTask
from .metadata import Metadata
from .resource import Resource
from .package import Package
from . import helpers
from . import config


class Pipeline(Metadata):
    """Pipeline representation.

    Parameters:
        descriptor? (str|dict): pipeline descriptor

    Raises:
        FrictionlessException: raise any error that occurs during the process

    """

    def __init__(self, descriptor, tasks=None):
        self.setinitial("tasks", tasks)
        super().__init__(descriptor)

    @property
    def tasks(self):
        """
        Returns:
            dict[]: tasks
        """
        tasks = self.get("tasks", [])
        return self.metadata_attach("tasks", tasks)

    # Run

    def run(self, *, parallel=False):
        """Run the pipeline"""
        tasks = []
        timer = helpers.Timer()
        for task in self.tasks:
            errors = []
            target = None
            try:
                target = task.run()
            except Exception as exception:
                errors.append(TaskError(note=str(exception)))
            tasks.append(StatusTask(errors=errors, target=target, type=task.type))
        return Status(tasks=tasks, time=timer.time, errors=[])

    # Metadata

    metadata_strict = True
    metadata_Error = PipelineError
    metadata_profile = config.PIPELINE_PROFILE

    def metadata_process(self):

        # Tasks
        tasks = self.get("tasks")
        if isinstance(tasks, list):
            for index, task in enumerate(tasks):
                if not isinstance(task, PipelineTask):
                    task = PipelineTask(task)
                    list.__setitem__(tasks, index, task)
            if not isinstance(tasks, helpers.ControlledList):
                tasks = helpers.ControlledList(tasks)
                tasks.__onchange__(self.metadata_process)
                dict.__setitem__(self, "tasks", tasks)


# NOTE: we can use metadata_process to task.source -> Resource/Package?
class PipelineTask(Metadata):
    """Pipeline task representation.

    Parameters:
        descriptor? (str|dict): pipeline task descriptor

    Raises:
        FrictionlessException: raise any error that occurs during the process

    """

    def __init__(self, descriptor=None, *, source=None, type=None, steps=None):
        self.setinitial("source", source)
        self.setinitial("type", type)
        self.setinitial("steps", steps)
        super().__init__(descriptor)

    @property
    def source(self):
        return self["source"]

    @property
    def type(self):
        return self["type"]

    @property
    def steps(self):
        return self["steps"]

    def run(self):
        """Run the task"""
        transforms = import_module("frictionless.transform")

        # Resource transform
        if self.type == "resource":
            source = Resource(self.source)
            return transforms.transform_resource(source, steps=self.steps)

        # Package transform
        elif self.type == "package":
            source = Package(self.source)
            return transforms.transform_package(source, steps=self.steps)

        # Not supported transform
        note = f'Transform type "{self.type}" is not supported'
        raise FrictionlessException(PipelineError(note=note))

    # Metadata

    metadata_strict = True
    metadata_Error = PipelineError
    metadata_profile = config.PIPELINE_PROFILE["properties"]["tasks"]["items"]



import sys
import typer
from typing import List
from typer import Option as Opt
from typer import Argument as Arg
from ..describe import describe
from ..detector import Detector
from ..layout import Layout
from .main import program
from .. import helpers


@program.command(name="describe")
def program_describe(
    source: List[str] = Arg(None, help="Data source to describe [default: stdin]"),
    type: str = Opt(None, help='Specify source type e.g. "package"'),
    # File
    scheme: str = Opt(None, help="Specify schema  [default: inferred]"),
    format: str = Opt(None, help="Specify format  [default: inferred]"),
    hashing: str = Opt(None, help="Specify hashing algorithm  [default: inferred]"),
    encoding: str = Opt(None, help="Specify encoding  [default: inferred]"),
    innerpath: str = Opt(None, help="Specify in-archive path  [default: first]"),
    compression: str = Opt(None, help="Specify compression  [default: inferred]"),
    # Table
    header_rows: str = Opt(None, help="Comma-separated row numbers  [default: 1]"),
    header_join: str = Opt(None, help="A separator to join a multiline header"),
    pick_fields: str = Opt(None, help='Comma-separated fields to pick e.g. "1,name1"'),
    skip_fields: str = Opt(None, help='Comma-separated fields to skip e.g. "2,name2"'),
    limit_fields: int = Opt(None, help="Limit fields by this integer"),
    offset_fields: int = Opt(None, help="Offset fields by this integer"),
    pick_rows: str = Opt(None, help='Comma-separated rows to pick e.g. "1,<blank>"'),
    skip_rows: str = Opt(None, help='Comma-separated rows to skip e.g. "2,3,4,5"'),
    limit_rows: int = Opt(None, help="Limit rows by this integer"),
    offset_rows: int = Opt(None, help="Offset rows by this integer"),
    # Detector
    buffer_size: int = Opt(None, help="Limit byte buffer size by this integer"),
    sample_size: int = Opt(None, help="Limit data sample size by this integer"),
    field_type: str = Opt(None, help="Force all the fields to have this type"),
    field_names: str = Opt(None, help="Comma-separated list of field names"),
    field_confidence: float = Opt(None, help="A float from 0 to 1"),
    field_float_numbers: bool = Opt(None, help="Make number floats instead of decimals"),
    field_missing_values: str = Opt(None, help="Comma-separated list of missing values"),
    # Misc
    basepath: str = Opt(None, help="Basepath of the resource/package"),
    expand: bool = Opt(None, help="Expand default values"),
    nostats: bool = Opt(None, help="Do not infer stats"),
    yaml: bool = Opt(False, help="Return in pure YAML format"),
    json: bool = Opt(False, help="Return in JSON format"),
):
    """
    Describe a data source.

    Based on the inferred data source type it will return resource or package descriptor.
    Default output format is YAML with a front matter.
    """

    # Support stdin
    is_stdin = False
    if not source:
        is_stdin = True
        source = [helpers.create_byte_stream(sys.stdin.buffer.read())]

    # Normalize parameters
    source = list(source) if len(source) > 1 else source[0]
    header_rows = helpers.parse_csv_string(header_rows, convert=int)
    pick_fields = helpers.parse_csv_string(pick_fields, convert=int, fallback=True)
    skip_fields = helpers.parse_csv_string(skip_fields, convert=int, fallback=True)
    pick_rows = helpers.parse_csv_string(pick_rows, convert=int, fallback=True)
    skip_rows = helpers.parse_csv_string(skip_rows, convert=int, fallback=True)
    field_names = helpers.parse_csv_string(field_names)
    field_missing_values = helpers.parse_csv_string(field_missing_values)

    # Prepare layout
    layout = (
        Layout(
            header_rows=header_rows,
            header_join=header_join,
            pick_fields=pick_fields,
            skip_fields=skip_fields,
            limit_fields=limit_fields,
            offset_fields=offset_fields,
            pick_rows=pick_rows,
            skip_rows=skip_rows,
            limit_rows=limit_rows,
            offset_rows=offset_rows,
        )
        or None
    )

    # Prepare detector
    detector = Detector(
        **helpers.remove_non_values(
            dict(
                buffer_size=buffer_size,
                sample_size=sample_size,
                field_type=field_type,
                field_names=field_names,
                field_confidence=field_confidence,
                field_float_numbers=field_float_numbers,
                field_missing_values=field_missing_values,
            )
        )
    )

    # Prepare options
    options = helpers.remove_non_values(
        dict(
            type=type,
            # Spec
            scheme=scheme,
            format=format,
            hashing=hashing,
            encoding=encoding,
            innerpath=innerpath,
            compression=compression,
            layout=layout,
            # Extra
            detector=detector,
            expand=expand,
            nostats=nostats,
        )
    )

    # Describe source
    try:
        metadata = describe(source, **options)
    except Exception as exception:
        typer.secho(str(exception), err=True, fg=typer.colors.RED, bold=True)
        raise typer.Exit(1)

    # Return JSON
    if json:
        descriptor = metadata.to_json()
        typer.secho(descriptor)
        raise typer.Exit()

    # Return YAML
    if yaml:
        descriptor = metadata.to_yaml().strip()
        typer.secho(descriptor)
        raise typer.Exit()

    # Return default
    if is_stdin:
        source = "stdin"
    elif isinstance(source, list):
        source = " ".join(source)
    typer.secho("---")
    typer.secho(f"metadata: {source}", bold=True)
    typer.secho("---")
    typer.secho("")
    typer.secho(metadata.to_yaml().strip())
    typer.secho("")



from .. import errors
from ..check import Check


class checksum(Check):
    """Check a table's checksum

    API      | Usage
    -------- | --------
    Public   | `from frictionless import checks`
    Implicit | `validate(checksum={...})`

    Ths check is enabled by default if the `checksum` argument
    is provided for the `validate` function.

    Parameters:
       descriptor (dict): check's descriptor
       hash? (str): a hash sum of the table's bytes
       bytes? (int): number of bytes
       fields? (int): number of fields
       rows? (int): number of rows

    """

    code = "checksum"
    Errors = [errors.ChecksumError]

    def __init__(self, descriptor=None, *, hash=None, bytes=None, fields=None, rows=None):
        self.setinitial("hash", hash)
        self.setinitial("bytes", bytes)
        self.setinitial("fields", fields)
        self.setinitial("rows", rows)
        super().__init__(descriptor)

    # Validate

    def validate_table(self):

        # Hash
        if self.get("hash"):
            hashing = self.resource.hashing
            if self["hash"] != self.resource.stats["hash"]:
                note = 'expected hash in %s is "%s" and actual is "%s"'
                note = note % (hashing, self["hash"], self.resource.stats["hash"])
                yield errors.ChecksumError(note=note)

        # Bytes
        if self.get("bytes"):
            if self["bytes"] != self.resource.stats["bytes"]:
                note = 'expected bytes count is "%s" and actual is "%s"'
                note = note % (self["bytes"], self.resource.stats["bytes"])
                yield errors.ChecksumError(note=note)

        # Fields
        if self.get("fields"):
            if self["fields"] != self.resource.stats["fields"]:
                note = 'expected fields count is "%s" and actual is "%s"'
                note = note % (self["fields"], self.resource.stats["fields"])
                yield errors.ChecksumError(note=note)

        # Rows
        if self.get("rows"):
            if self["rows"] != self.resource.stats["rows"]:
                note = 'expected rows count is "%s" and actual is "%s"'
                note = note % (self["rows"], self.resource.stats["rows"])
                yield errors.ChecksumError(note=note)

    # Metadata

    metadata_profile = {  # type: ignore
        "type": "object",
        "properties": {
            "hash": {"type": "string"},
            "bytes": {"type": "number"},
            "fields": {"type": "number"},
            "rows": {"type": "number"},
        },
    }



from ..check import Check
from .. import errors


class baseline(Check):
    """Check a table for basic errors

    API      | Usage
    -------- | --------
    Public   | `from frictionless import checks`
    Implicit | `validate(...)`

    Ths check is enabled by default for any `validate` function run.

    """

    code = "baseline"
    Errors = [
        # Header
        errors.BlankHeaderError,
        # Label
        errors.ExtraLabelError,
        errors.MissingLabelError,
        errors.BlankLabelError,
        errors.DuplicateLabelError,
        errors.IncorrectLabelError,
        # Row
        errors.BlankRowError,
        errors.PrimaryKeyError,
        errors.ForeignKeyError,
        # Cell
        errors.ExtraCellError,
        errors.MissingCellError,
        errors.TypeError,
        errors.ConstraintError,
        errors.UniqueError,
    ]

    # Validate

    def validate_source(self):
        empty = not (self.resource.sample or self.resource.labels)
        yield from [errors.SourceError(note="the source is empty")] if empty else []

    def validate_schema(self):
        empty = not (self.resource.sample or self.resource.labels)
        yield from self.resource.schema.metadata_errors if not empty else []

    def validate_header(self):
        yield from self.resource.header.errors

    def validate_row(self, row):
        yield from row.errors

    # Metadata

    metadata_profile = {  # type: ignore
        "type": "object",
        "properties": {},
    }



import pytest
from frictionless import Resource, transform, steps


# General


@pytest.mark.skip
def test_transform_resource():
    source = Resource(path="data/transform.csv")
    source.infer()
    target = transform(
        source,
        steps=[
            steps.table_normalize(),
            steps.table_melt(field_name="id"),
            steps.table_recast(field_name="id"),
        ],
    )
    assert target.schema == source.schema
    assert target.read_rows() == [
        {"id": 1, "name": "germany", "population": 83},
        {"id": 2, "name": "france", "population": 66},
        {"id": 3, "name": "spain", "population": 47},
    ]


@pytest.mark.skip
def test_transform_resource_from_descriptor():
    source = Resource(path="data/transform.csv")
    source.infer()
    target = transform(
        source.to_dict(),
        steps=[
            steps.table_normalize(),
            steps.table_melt(field_name="id"),
            steps.table_recast(field_name="id"),
        ],
    )
    assert target.schema == source.schema
    assert target.read_rows() == [
        {"id": 1, "name": "germany", "population": 83},
        {"id": 2, "name": "france", "population": 66},
        {"id": 3, "name": "spain", "population": 47},
    ]



import pytest
from frictionless import Resource, transform, steps


# General


@pytest.mark.skip
def test_transform():
    source = Resource(path="data/transform.csv")
    source.infer()
    target = transform(
        source,
        steps=[
            steps.table_normalize(),
            steps.table_melt(field_name="id"),
            steps.table_recast(field_name="id"),
        ],
    )
    assert target.schema == source.schema
    assert target.read_rows() == [
        {"id": 1, "name": "germany", "population": 83},
        {"id": 2, "name": "france", "population": 66},
        {"id": 3, "name": "spain", "population": 47},
    ]


def test_transform_custom_step_function_based():

    # Create step
    def custom(resource):
        with resource:
            for row in resource.row_stream:
                row["id"] = row["id"] * row["id"]
                yield row

    # Transform resource
    source = Resource(path="data/transform.csv")
    source.infer()
    target = transform(source, steps=[custom])
    assert target.schema == source.schema
    assert target.read_rows() == [
        {"id": 1, "name": "germany", "population": 83},
        {"id": 4, "name": "france", "population": 66},
        {"id": 9, "name": "spain", "population": 47},
    ]



import types
from ..check import Check
from ..system import system
from ..resource import Resource
from ..exception import FrictionlessException
from ..report import Report, ReportTask
from .. import config
from .. import errors
from .. import helpers


@Report.from_validate
def validate_resource(
    source,
    *,
    # Validation
    checks=None,
    checksum=None,
    pick_errors=None,
    skip_errors=None,
    limit_errors=config.DEFAULT_LIMIT_ERRORS,
    limit_memory=config.DEFAULT_LIMIT_MEMORY,
    noinfer=False,
    **options,
):
    """Validate table

    API      | Usage
    -------- | --------
    Public   | `from frictionless import validate_table`

    Parameters:
        source (any): the source of the resource
        checks? (list): a list of checks
        checksum? (dict): a checksum dictionary
        pick_errors? ((str|int)[]): pick errors
        skip_errors? ((str|int)[]): skip errors
        limit_errors? (int): limit errors
        limit_memory? (int): limit memory
        noinfer? (bool): validate resource as it is
        **options? (dict): Resource constructor options

    Returns:
        Report: validation report
    """

    # Create state
    partial = False
    task_errors = []
    table_errors = TableErrors(pick_errors, skip_errors, limit_errors)
    timer = helpers.Timer()

    # Create resource
    try:
        native = isinstance(source, Resource)
        resource = source.to_copy() if native else Resource(source, **options)
    except FrictionlessException as exception:
        return Report(time=timer.time, errors=[exception.error], tasks=[])

    # Prepare checksum
    if not checksum:
        checksum = {key: value for key, value in resource.stats.items() if value}

    # Prepare resource
    if not noinfer:
        resource.infer()
    if resource.metadata_errors:
        return Report(time=timer.time, errors=resource.metadata_errors, tasks=[])

    # Open resource
    try:
        resource.open()
    except FrictionlessException as exception:
        table_errors.append(exception.error, force=True)
        resource.close()

    # Prepare checks
    checks = checks or []
    checks.insert(0, {"code": "baseline"})
    if checksum:
        checks.insert(1, {"code": "checksum", **checksum})
    for index, check in enumerate(checks):
        if not isinstance(check, Check):
            checks[index] = (
                Check(function=check)
                if isinstance(check, types.FunctionType)
                else system.create_check(check)
            )

    # Enter table
    if not table_errors:
        with resource:

            # Prepare checks
            for check in checks:
                table_errors.register(check)
                check.connect(resource)
                check.prepare()

            # Validate checks
            for check in checks.copy():
                for error in check.validate_check():
                    task_errors.append(error)
                    if check in checks:
                        checks.remove(check)

            # Validate source
            for check in checks:
                for error in check.validate_source():
                    table_errors.append(error, force=True)

            # Validate schema
            for check in checks:
                for error in check.validate_schema():
                    table_errors.append(error, force=True)

            # Validate header
            if resource.header:
                for check in checks:
                    for error in check.validate_header():
                        table_errors.append(error)

            # Validate rows
            for row in resource.row_stream:

                # Validate row
                for check in checks:
                    for error in check.validate_row(row):
                        table_errors.append(error)

                # Limit errors
                if limit_errors and len(table_errors) >= limit_errors:
                    partial = True
                    break

                # Limit memory
                if limit_memory and not resource.stats["rows"] % 100000:
                    memory = helpers.get_current_memory_usage()
                    if memory and memory > limit_memory:
                        note = f'exceeded memory limit "{limit_memory}MB"'
                        task_errors.append(errors.TaskError(note=note))
                        partial = True
                        break

            # Validate table
            if not partial:
                for check in checks:
                    for error in check.validate_table():
                        table_errors.append(error)

    # Return report
    return Report(
        time=timer.time,
        errors=task_errors,
        tasks=[
            ReportTask(
                time=timer.time,
                scope=table_errors.scope,
                partial=partial,
                errors=table_errors,
                resource=resource,
            )
        ],
    )


# Internal


class TableErrors(list):
    def __init__(self, pick_errors, skip_errors, limit_errors):
        self.__pick_errors = set(pick_errors or [])
        self.__skip_errors = set(skip_errors or [])
        self.__limit_errors = limit_errors
        self.__scope = []

    @property
    def scope(self):
        return self.__scope

    def append(self, error, *, force=False):
        if not force:
            if self.__limit_errors:
                if len(self) >= self.__limit_errors:
                    return
            if not self.match(error):
                return
        super().append(error)

    def match(self, error):
        match = True
        if self.__pick_errors:
            match = False
            if error.code in self.__pick_errors:
                match = True
            if self.__pick_errors.intersection(error.tags):
                match = True
        if self.__skip_errors:
            match = True
            if error.code in self.__skip_errors:
                match = False
            if self.__skip_errors.intersection(error.tags):
                match = False
        return match

    def register(self, check):
        for Error in check.Errors:
            if not self.match(Error):
                continue
            if Error.code in self.__scope:
                continue
            self.__scope.append(Error.code)



from .baseline import baseline
from .checksum import checksum
from .heuristic import duplicate_row, deviated_value, truncated_value
from .regulation import forbidden_value, sequential_value, row_constraint



class Storage:
    """Storage representation"""

    def __init__(self, source, **options):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    # Read

    def read_resource(self, name, **options):
        raise NotImplementedError()

    def read_package(self, **options):
        raise NotImplementedError()

    # Write

    def write_resource(self, resource, *, force=False, **options):
        raise NotImplementedError()

    def write_package(self, package, *, force=False, **options):
        raise NotImplementedError()

    # Delete

    def delete_resource(self, name, *, ignore=False, **options):
        raise NotImplementedError()

    def delete_package(self, names, *, ignore=False, **options):
        raise NotImplementedError()



from .. import helpers
from ..report import Report
from ..package import Package
from ..inquiry import Inquiry, InquiryTask
from ..exception import FrictionlessException
from .inquiry import validate_inquiry


@Report.from_validate
def validate_package(source, noinfer=False, nolookup=False, parallel=False, **options):
    """Validate package

    API      | Usage
    -------- | --------
    Public   | `from frictionless import validate_package`

    Parameters:
        source (dict|str): a package descriptor
        basepath? (str): package basepath
        trusted? (bool): don't raise an exception on unsafe paths
        noinfer? (bool): don't call `package.infer`
        nolookup? (bool): don't read lookup tables skipping integrity checks
        parallel? (bool): enable multiprocessing
        **options (dict): Package constructor options

    Returns:
        Report: validation report

    """

    # Create state
    timer = helpers.Timer()

    # Create package
    try:
        native = isinstance(source, Package)
        package = source.to_copy() if native else Package(source, **options)
    except FrictionlessException as exception:
        return Report(time=timer.time, errors=[exception.error], tasks=[])

    # Prepare package
    if not noinfer:
        package.infer()

    if package.metadata_errors:
        return Report(time=timer.time, errors=package.metadata_errors, tasks=[])

    # Prepare inquiry
    inquiry = Inquiry(tasks=[])
    for resource in package.resources:
        # NOTE: we should consider validating non tabular resources either
        if resource.profile == "tabular-data-resource":
            inquiry.tasks.append(
                InquiryTask(
                    source=resource,
                    basepath=resource.basepath,
                    noinfer=noinfer,
                )
            )

    # Validate inquiry
    report = validate_inquiry(inquiry, parallel=parallel)

    # Return report
    return Report(time=timer.time, errors=report["errors"], tasks=report["tasks"])



from .exception import FrictionlessException
from .errors import Error, StatusError
from .metadata import Metadata
from . import config


class Status(Metadata):
    """Status representation.

    Parameters:
        descriptor? (str|dict): schema descriptor

    Raises:
        FrictionlessException: raise any error that occurs during the process

    """

    def __init__(self, descriptor=None, *, time, errors, tasks):
        stats_errors = len(errors) + sum(task["stats"]["errors"] for task in tasks)
        self.setinitial("version", config.VERSION)
        self.setinitial("time", time)
        self.setinitial("valid", not errors and all(task["valid"] for task in tasks))
        self.setinitial("stats", {"errors": stats_errors, "tasks": len(tasks)})
        self.setinitial("errors", errors)
        self.setinitial("tasks", tasks)
        super().__init__(descriptor)

    @property
    def version(self):
        """
        Returns:
            str: frictionless version
        """
        return self["version"]

    @property
    def time(self):
        """
        Returns:
            float: validation time
        """
        return self["time"]

    @property
    def valid(self):
        """
        Returns:
            bool: validation result
        """
        return self["valid"]

    @property
    def stats(self):
        """
        Returns:
            dict: validation stats
        """
        return self["stats"]

    @property
    def errors(self):
        """
        Returns:
            Error[]: validation errors
        """
        return self["errors"]

    @property
    def tasks(self):
        """
        Returns:
            ReportTable[]: validation tasks
        """
        return self["tasks"]

    @property
    def task(self):
        """
        Returns:
            ReportTable: validation task (if there is only one)

        Raises:
            FrictionlessException: if there are more that 1 task
        """
        if len(self.tasks) != 1:
            error = Error(note='The "report.task" is available for single task reports')
            raise FrictionlessException(error)
        return self.tasks[0]

    # Metadata

    metadata_strict = True
    metadata_Error = StatusError
    metadata_profile = config.STATUS_PROFILE


class StatusTask(Metadata):
    """ Status Task representation"""

    def __init__(self, descriptor=None, *, errors, target, type):
        self.setinitial("valid", not errors)
        self.setinitial("errors", errors)
        self.setinitial("target", target)
        self.setinitial("type", type)
        self.setinitial("stats", {"errors": len(errors)})
        super().__init__(descriptor)

    @property
    def valid(self):
        """
        Returns:
            bool: validation result
        """
        return self["valid"]

    @property
    def errors(self):
        """
        Returns:
            Error[]: validation errors
        """
        return self["errors"]

    @property
    def target(self):
        """
        Returns:
            any: validation target
        """
        return self["target"]

    @property
    def type(self):
        """
        Returns:
            any: validation target
        """
        return self["type"]

    # Metadata

    metadata_strict = True
    metadata_Error = StatusError
    metadata_profile = config.STATUS_PROFILE["properties"]["tasks"]["items"]



from ..plugin import Plugin
from ..loader import Loader
from ..control import Control
from ..exception import FrictionlessException
from .. import errors


# Plugin


class StreamPlugin(Plugin):
    """Plugin for Local Data

    API      | Usage
    -------- | --------
    Public   | `from frictionless.plugins.stream import StreamPlugin`

    """

    def create_file(self, file):
        if not file.scheme and not file.format:
            if hasattr(file.data, "read"):
                file.scheme = "stream"
                file.format = ""
                return file

    def create_control(self, resource, *, descriptor):
        if resource.scheme == "stream":
            return StreamControl(descriptor)

    def create_loader(self, resource):
        if resource.scheme == "stream":
            return StreamLoader(resource)


# Control


class StreamControl(Control):
    """Stream control representation

    API      | Usage
    -------- | --------
    Public   | `from frictionless.plugins.stream import StreamControl`

    Parameters:
        descriptor? (str|dict): descriptor

    Raises:
        FrictionlessException: raise any error that occurs during the process

    """

    pass


# Loader


class StreamLoader(Loader):
    """Stream loader implementation.

    API      | Usage
    -------- | --------
    Public   | `from frictionless.plugins.stream import StreamLoader`

    """

    # Read

    def read_byte_stream_create(self):
        data = self.resource.data
        if hasattr(data, "encoding"):
            error = errors.SchemeError(note="only byte streams are supported")
            raise FrictionlessException(error)
        return data

    # Write

    def write_byte_stream_save(self, byte_stream):
        return byte_stream



import stringcase
from functools import partial
from multiprocessing import Pool
from importlib import import_module
from .metadata import Metadata
from .exception import FrictionlessException
from .errors import InquiryError
from .system import system
from .report import Report
from . import helpers
from . import config


class Inquiry(Metadata):
    """Inquiry representation.

    Parameters:
        descriptor? (str|dict): descriptor

    Raises:
        FrictionlessException: raise any error that occurs during the process

    """

    def __init__(self, descriptor=None, *, tasks=None):
        self.setinitial("tasks", tasks)
        super().__init__(descriptor)

    @property
    def tasks(self):
        """
        Returns:
            dict[]: tasks
        """
        return self["tasks"]

    # Run

    def run(self, *, parallel=False):
        validate = import_module("frictionless").validate

        # Create state
        tasks = []
        reports = []
        timer = helpers.Timer()

        # Prepare tasks
        for task in self.tasks:
            if task.type == "inquiry":
                note = "Inquiry cannot contain nested inquiries"
                raise FrictionlessException(InquiryError(note=note))
            if task.type == "package":
                # NOTE:
                # For now, we don't flatten inquiry completely and for the case
                # of a list of packages with one resource we don't get proper multiprocessing
                report = validate(**helpers.create_options(task))
                reports.append(report)
                continue
            tasks.append(task)

        # Validate sequentially
        if len(tasks) == 1 or not parallel:
            for task in tasks:
                report = validate(**helpers.create_options(task))
                reports.append(report)

        # Validate in-parallel
        else:
            with Pool() as pool:
                # NOTE:
                # We need more tests to guarantee proper serialization
                # In general, we need to send plain object and get plain objects back
                reports.extend(pool.map(partial(helpers.apply_function, validate), tasks))

        # Return report
        errors = []
        tasks = []
        for report in reports:
            errors.extend(report["errors"])
            tasks.extend(report["tasks"])
        return Report(time=timer.time, errors=errors, tasks=tasks)

    # Metadata

    metadata_strict = True
    metadata_Error = InquiryError
    metadata_profile = config.INQUIRY_PROFILE

    def metadata_process(self):

        # Tasks
        tasks = self.get("tasks")
        if isinstance(tasks, list):
            for index, task in enumerate(tasks):
                if not isinstance(task, InquiryTask):
                    task = InquiryTask(task)
                    list.__setitem__(tasks, index, task)
            if not isinstance(tasks, helpers.ControlledList):
                tasks = helpers.ControlledList(tasks)
                tasks.__onchange__(self.metadata_process)
                dict.__setitem__(self, "tasks", tasks)


class InquiryTask(Metadata):
    """Inquiry task representation.

    Parameters:
        descriptor? (str|dict): descriptor

    Raises:
        FrictionlessException: raise any error that occurs during the process

    """

    def __init__(self, descriptor=None, *, source=None, type=None, **options):
        self.setinitial("source", source)
        self.setinitial("type", type)
        for key, value in options.items():
            self.setinitial(stringcase.camelcase(key), value)
        super().__init__(descriptor)

    @property
    def source(self):
        """
        Returns:
            any: source
        """
        return self["source"]

    @property
    def type(self):
        """
        Returns:
            string?: type
        """
        return self.get("type") or system.create_file(self.source).type

    # Metadata

    metadata_strict = True
    metadata_Error = InquiryError
    metadata_profile = config.INQUIRY_PROFILE["properties"]["tasks"]["items"]
